/* ----------------------------------------------------------------------------
 *  gpsWrapper
 * ----------------------------------------------------------------------------
 *  just a wrapper class around TinyGps + some Functions from Wirering NMEA-Lib
 * ----------------------------------------------------------------------------
 *  prog: max.rheiner@zhdk.ch
 * ----------------------------------------------------------------------------
 */

#include "gpsWrapper.h"

bool GpsWrapper::start(HardwareSerial* serialPort,int baudRate)
{
  if(serialPort == NULL)
    return false;

  _serial = serialPort;
  _serial->begin(baudRate);
  return true;
}

void GpsWrapper::end()
{
  if(_serial == NULL)
    return;
}

bool GpsWrapper::parseGps()
{
  while(_serial->available())
  {
    if( _gps.encode(_serial->read()) )
    { // check if there is valid data
      updateData();
      return true;
    }
  }
  return false;
}

void GpsWrapper::updateData()
{
  _gps.get_position(&_lat, &_long,&_fix);
  if(_fix == TinyGPS::GPS_INVALID_AGE) 
  { // nothing to read, no gps data 
    _valid= false;
    return;
  }
  
  // calc date
  _gps.crack_datetime(&_year,&_month,&_day,&_hour,&_minute,&_second,&_hundredths,&_fix);
}

void GpsWrapper::dump(HardwareSerial* serialPort)
{
  long lat, lon;
  float flat, flon;
  unsigned long age, date, time, chars;
  int year;
  byte month, day, hour, minute, second, hundredths;
  unsigned short sentences, failed;

  serialPort->println("--------------------------------");
 
  _gps.get_position(&lat, &lon, &age);
  if(age == TinyGPS::GPS_INVALID_AGE)
  {
    serialPort->println("No valid data !");
    return;
  }
  serialPort->print("Lat/Long(10^-5 deg): "); serialPort->print(lat); serialPort->print(", "); serialPort->print(lon); 
  serialPort->print(" Fix age: "); serialPort->print(age); serialPort->println("ms.");
  
  parseGps(); // If we don't feed the gps during this long routine, we may drop characters and get checksum errors

  _gps.f_get_position(&flat, &flon, &age);
  serialPort->print("Lat/Long(float): "); printFloat(flat, 5); serialPort->print(", "); printFloat(flon, 5);
  serialPort->print(" Fix age: "); serialPort->print(age); serialPort->println("ms.");

  parseGps();

  _gps.get_datetime(&date, &time, &age);
  serialPort->print("Date(ddmmyy): "); serialPort->print(date); serialPort->print(" Time(hhmmsscc): "); serialPort->print(time);
  serialPort->print(" Fix age: "); serialPort->print(age); serialPort->println("ms.");

  parseGps();

  _gps.crack_datetime(&year, &month, &day, &hour, &minute, &second, &hundredths, &age);
  serialPort->print("Date: "); serialPort->print(static_cast<int>(month)); serialPort->print("/"); serialPort->print(static_cast<int>(day)); serialPort->print("/"); serialPort->print(year);
  serialPort->print("  Time: "); serialPort->print(static_cast<int>(hour)); serialPort->print(":"); serialPort->print(static_cast<int>(minute)); serialPort->print(":"); serialPort->print(static_cast<int>(second)); serialPort->print("."); serialPort->print(static_cast<int>(hundredths));
  serialPort->print("  Fix age: ");  serialPort->print(age); serialPort->println("ms.");
  
  parseGps();

  serialPort->print("Alt(cm): "); serialPort->print(_gps.altitude()); serialPort->print(" Course(10^-2 deg): "); serialPort->print(_gps.course()); serialPort->print(" Speed(10^-2 knots): "); serialPort->println(_gps.speed());
  serialPort->print("Alt(float): "); printFloat(_gps.f_altitude()); serialPort->print(" Course(float): "); printFloat(_gps.f_course()); serialPort->println();
  serialPort->print("Speed(knots): "); printFloat(_gps.f_speed_knots()); serialPort->print(" (mph): ");  printFloat(_gps.f_speed_mph());
  serialPort->print(" (mps): "); printFloat(_gps.f_speed_mps()); serialPort->print(" (kmph): "); printFloat(_gps.f_speed_kmph()); serialPort->println();

  parseGps();

  _gps.stats(&chars, &sentences, &failed);
  serialPort->print("Stats: characters: "); serialPort->print(chars); serialPort->print(" sentences: "); serialPort->print(sentences); serialPort->print(" failed checksum: "); serialPort->println(failed);
 }

//////////////////////////////////////////////////////////////////////////////////////////////////
// static methods, took code from Wiring-NMEA lib

float GpsWrapper::distance(float lat1, float long1, float lat2, float long2, float units_per_meter)
{
	// returns distance in meters between two positions, both specified
	// as signed decimal-degrees latitude and longitude. Uses great-circle
	// distance computation for hypothised sphere of radius 6372795 meters.
	// Because Earth is no exact sphere, rounding errors may be upto 0.5%.
  float delta = radians(long1-long2);
  float sdlong = sin(delta);
  float cdlong = cos(delta);
  lat1 = radians(lat1);
  lat2 = radians(lat2);
  float slat1 = sin(lat1);
  float clat1 = cos(lat1);
  float slat2 = sin(lat2);
  float clat2 = cos(lat2);
  delta = (clat1 * slat2) - (slat1 * clat2 * cdlong);
  delta = sq(delta);
  delta += sq(clat2 * sdlong);
  delta = sqrt(delta);
  float denom = (slat1 * slat2) + (clat1 * clat2 * cdlong);
  delta = atan2(delta, denom);
  return delta * 6372795 * units_per_meter;
}

float GpsWrapper::angle(float lat1, float long1, float lat2, float long2) 
{
	// returns initial course in degrees (North=0, West=270) from
	// position 1 to position 2, both specified as signed decimal-degrees
	// latitude and longitude.
  float dlon = radians(long2-long1);
  lat1 = radians(lat1);
  lat2 = radians(lat2);
  float a1 = sin(dlon) * cos(lat2);
  float a2 = sin(lat1) * cos(lat2) * cos(dlon);
  a2 = cos(lat1) * sin(lat2) - a2;
  a2 = atan2(a1, a2);
  if (a2 < 0.0) {
  	a2 += TWO_PI;			// modulo operator doesn't seem to work on floats
  }
  return degrees(a2);
}

void GpsWrapper::printFloat(float number, int digits)
{
  // Handle negative numbers
  if (number < 0.0)
  {
     Serial.print('-');
     number = -number;
  }

  // Round correctly so that print(1.999, 2) prints as "2.00"
  float rounding = 0.5;
  for (uint8_t i=0; i<digits; ++i)
    rounding /= 10.0;
  
  number += rounding;

  // Extract the integer part of the number and print it
  unsigned long int_part = (unsigned long)number;
  float remainder = number - (float)int_part;
  Serial.print(int_part);

  // Print the decimal point, but only if there are digits beyond
  if (digits > 0)
    Serial.print("."); 

  // Extract digits from the remainder one at a time
  while (digits-- > 0)
  {
    remainder *= 10.0;
    int toPrint = int(remainder);
    Serial.print(toPrint);
    remainder -= toPrint; 
  } 
}

